#!/usr/bin/perl -w
#
# shinGETsu - P2P anonymous BBS
#
# (c) Fuktommy <fuktommy@users.sourceforge.net>
# Released under the GNU General Public License
# $Id: httpd.pl,v 1.45 2004/04/27 12:11:15 fuktommy Exp $
#

use lib qw(.);
use Socket;
use IPC::Open2;
use Shingetsu::Config;
use Shingetsu::Extern;
use Shingetsu::Util;

sub xdie(@)   { killClient(); die @_; }
sub dprint(@) { print STDOUT @_, "\n" if ($debug); }
sub xlocaltime(;$) { Shingetsu::Util::localtime(@_) }
sub killClient {
	if ((defined $pid_client)&&($pid_client>0)) {
		kill "TERM", $pid_client;
		dprint "terminate $pid_client";
	}
	unlink $Shingetsu::Config::pid;
}

$SIG{CHLD} = "IGNORE";
$SIG{TERM} = $SIG{HUP} = sub {
	killClient();
	exit;
};

#
# default value
#
$port	 = $Shingetsu::Config::port;
$debug	 = $Shingetsu::Config::debug;
$cgiDir	 = $Shingetsu::Config::cgiDir;
$timeout = $Shingetsu::Config::timeout;
$bufsize = $Shingetsu::Config::bufsize;
@initNode = ();

#
# arguments
#
while (@ARGV) {
	$_ = shift @ARGV;
	if ($_ eq "-v") {
		$debug = 1;
	} elsif ($_ eq "-p") {
		$port = shift @ARGV;
	} elsif ($_ eq "-i") {
		push @initNode, shift(@ARGV);
	}
}
chdir $cgiDir or die "$cgiDir: no such directory. $!\n";
unlink $Shingetsu::Config::client;
unless ($debug) {
	my $pid;
	if ($pid = fork) {
		$SIG{TERM} = $SIG{HUP} = "DEFAULT";
		open OUT, "> $Shingetsu::Config::pid";
		print OUT "$pid\n";
		close OUT;
		exit;
	}
	open STDIN,  "< /dev/null";
	open STDOUT, "> /dev/null";
	open STDERR, "> /dev/null";
}

#
# environment variable
#
$ENV{SERVER_SOFTWARE}	= "shinGETsu/$Shingetsu::Config::version";
$ENV{GATEWAY_INTERFACE} = "CGI/1.1";
$ENV{SERVER_PROTOCOL}	= "HTTP/1.0";
$ENV{SERVER_PORT}	= $port;
$ENV{HTTP_PORT}		= $port;
$ENV{SERVER_NAME}	= "";
$ENV{HTTP_HOST}		= "";
$ENV{X_INIT_NODE}	= join "+", @initNode if (@initNode);

#
# client process
#
if (!($pid_client = fork)) {
	client();
}

#
# daemon process
#
httpd($port, $cgiDir);

#------------------------------------------------------------------#

#
# httpd
# args: port, cgi-directory-path
#
sub httpd {
	my($port, $cgiDir) = @_;

	# initialize daemon
	socket CLIENT_WAITING, PF_INET, SOCK_STREAM, 0 or xdie "failed to open socket. $!\n";
	setsockopt CLIENT_WAITING, SOL_SOCKET, SO_REUSEADDR, 1 or xdie "failed to setsockopt. $!\n";
	bind CLIENT_WAITING, pack_sockaddr_in($port, INADDR_ANY) or xdie "failed to bind. $!\n";
	listen CLIENT_WAITING, SOMAXCONN or xdie "failed to listen: $!\n";

	# run daemon
	while (1) {
		sleep 1 if (rand() < 1/$Shingetsu::Config::childrenLimit);
		unless ($paddr = accept(CLIENT, CLIENT_WAITING)) {
			#dprint "faild to accept";
			sleep 1;
			next;
		}
		my ($client_port, $client_iaddr) = unpack_sockaddr_in $paddr;
		local $client_ip = inet_ntoa $client_iaddr;
		my $client_hostname;
		if ($Shingetsu::Config::resolv) {
			$client_hostname = gethostbyaddr($client_iaddr, AF_INET) || $client_ip;
		} else {
			$client_hostname = $client_ip;
		}

		select CLIENT; $|=1;

		if (! defined($pid = fork)) {
			dprint "server process: fork failed.";
			close CLIENT;
			next;

		} elsif ($pid) {
			close CLIENT;
			next;

		} else {
			$SIG{TERM} = $SIG{HUP} = $SIG{CHLD} = "DEFAULT";
			$SIG{ALRM} = sub { die "$client_ip: server timeout" };
			$ENV{REMOTE_HOST} = $client_hostname;
			$ENV{REMOTE_ADDR} = $client_ip;
			$ENV{REMOTE_PORT} = $client_port;

			# read HTTP request
			my($str_request, $referer, $agent) = ("", "", "");
			alarm $timeout;
			while (<CLIENT>) {
				s/[\n|\r]//g;
				if ($_ eq "") {
					last;

				} elsif (/^(\S+)\s+(\S+)/ && (!defined $ENV{REQUEST_METHOD})) {
					$str_request = $_;
					$ENV{REQUEST_METHOD} = $1;
					$ENV{REQUEST_URI} = $2;
					$ENV{REQUEST_URI} =~ s|^//+|/|g;

				} elsif (/^Host:\s+/i) {
					$ENV{HTTP_HOST} = $';

				} elsif ((/^Accept-Encoding:.*gzip/i)
						&& ($ENV{REMOTE_ADDR} !~ /$Shingetsu::Config::adminAddr/)) {
					$gzip = 1;

				} elsif (/^Accept-Language:\s+/) {
					$ENV{HTTP_ACCEPT_LANGUAGE} = $';

				} elsif (/^Range:\s*bytes\s*=\s*(\d+)\s*-/) {
					$range = $1;

				} elsif (/^Content-length:\s+(\d+)/i) {
					$ENV{CONTENT_LENGTH} = $1;

				} elsif (/^Referer:\s+/i) {
					$ENV{HTTP_REFERER} = $';
					$referer = $';

				} elsif (/^User-Agent:\s+/i) {
					$ENV{HTTP_USER_AGENT} = $';
					$agent = $';

				} elsif (/^Content-Type:\s+([^;]+)/i) {
					$ENV{CONTENT_TYPE} = $1;
				}
			}
			alarm 0;
			my $buf = "";
			if (! (defined $ENV{CONTENT_LENGTH})) {
			} elsif ($ENV{CONTENT_LENGTH} > $Shingetsu::Config::filelimit*1024*1024) {
				$flag_error = 1;
			} else {
				alarm $timeout;
				my $i;
				for ($i=0; $i<int($ENV{CONTENT_LENGTH}/$bufsize); $i++) {
					read CLIENT, $buf2, $bufsize;
					$buf .= $buf2;
					alarm $timeout;
				}
				if ($ENV{CONTENT_LENGTH} % $bufsize) {
					read CLIENT, $buf2, $ENV{CONTENT_LENGTH} % $bufsize;
					$buf .= $buf2;
				}
				alarm 0;
			}

			if ($ENV{REQUEST_URI} =~ m|^[a-z]+://([^/]+)|) {
				$ENV{HTTP_HOST}   = $1;
				$ENV{REQUEST_URI} = $';
			}
			if ($ENV{REQUEST_URI} eq "") {
				$ENV{REQUEST_URI} = "/";
			}

			if (($ENV{REQUEST_URI} =~ m|^/([0-9a-zA-Z]+)\.cgi|i)
				&& (-x "$1.cgi")) {
				$ENV{SCRIPT_NAME}  = "/$1.cgi";
				$ENV{PATH_INFO}	   = $';
				if ($ENV{PATH_INFO} =~ /\?/) {
					$ENV{QUERY_STRING} = $';
					$ENV{PATH_INFO}	  = $`;
				} else {
					$ENV{QUERY_STRING} = "";
				}
				
			} elsif ($ENV{REQUEST_URI} eq "/") {
				$ENV{SCRIPT_NAME}  = $Shingetsu::Config::index;
				$ENV{QUERY_STRING} = "";

			}

			# output logs
			my $date = xlocaltime;
			dprint "$client_ip<>$date<>$str_request<>$referer<>$agent";

			$gzip = 0 if ($ENV{SCRIPT_NAME}
				&& ($ENV{SCRIPT_NAME} ne "/server.cgi")
				&& ($ENV{SCRIPT_NAME} ne "/hosting.cgi"));

			if ($flag_error) {
				printHeader("404 Not Found");
				print "\r\nBad arguments or No data.\n",
				      "Try later.\n";

			} elsif ($ENV{REQUEST_METHOD} eq "OPTIONS") {
				printHeader("200 OK");
				print "Allow: GET, HEAD, POST, OPTIONS\n",
				      "Content-Length: 0\n\n";

			} elsif ($ENV{REQUEST_METHOD} !~ /^(GET|HEAD|POST)$/) {
				printHeader("501 Not Implemented");

			} elsif (defined $ENV{SCRIPT_NAME}) {
				$ENV{SCRIPT_FILENAME} = "$cgiDir$ENV{SCRIPT_NAME}";
				unless (open2(IN, OUT, ".$ENV{SCRIPT_NAME}")) {
					printHeader("404 Not Found");
					print "\r\nBad arguments or No data.\n",
					      "Try later.\n";
				} else {
					if ($ENV{CONTENT_LENGTH}) {
						print OUT $buf;
					}
					close OUT;
					my @buf = <IN>;
					close IN;
					wait;
					printResult(@buf);
				}

			} elsif ($ENV{REQUEST_URI} =~ m|^/[0-9A-Za-z._]+$|) {
				my $file = $&;
				printFile($file);

			} else {
				printHeader("404 Not Found");
				print "\r\nBad arguments or No data.\n",
				      "Try later.\n";
			}

			shutdown CLIENT, 2;
			close CLIENT;
			exit;
		}
	}
}

#------------------------------------------------------------------#

#
# print HTTP header
#
sub printHeader {
	my ($stat) = @_;
	my $ver = $Shingetsu::Config::version;
	local $, = "\r\n";
	alarm $timeout;
	print "HTTP/1.0 $stat",
	      "Server: shinGETsu/$ver",
	      "Connection: Close",
	      "";
	alarm 0;
}

#
# client process
#
sub client {
	$SIG{TERM} = $SIG{HUP} = "DEFAULT";
	$SIG{CHLD} = "IGNORE";
	sleep 5;
	while (1) {
		my $pid;
		if (! defined ($pid = fork)) {
			dprint "client process: fork failed.";
			sleep $Shingetsu::Config::clientfrec * 60;
			next;

		} elsif ($pid) {
			sleep 10;
			my $t = $Shingetsu::Config::clientfrec * 60
			      + int($Shingetsu::Config::timeError * Shingetsu::Util::normal());
			$t = $Shingetsu::Config::clientfrec * 60 if ($t <= 0);
			sleep $t;
			# wait; # ?????
			next;

		} else {
			open STDOUT, "> /dev/null";
			$ENV{QUERY_STRING}   = "";
			$ENV{PATH_INFO}	     = "";
			$ENV{REQUEST_METHOD} = "GET";
			$ENV{REMOTE_ADDR}    = "127.0.0.1";
			$ENV{SCRIPT_NAME}    = "/client.cgi";
			$SIG{CHLD} = "DEFAULT";
			exec "./client.cgi";
		}
	}
}

#
# print cgi result
#
sub printResult {
	my @buf = @_;
	my @head = ();
	while ($buf[0] !~ /^\s*$/) {
		push @head, shift(@buf);
	}
	shift @buf;
	my $buf = join "", @buf;
	my $len = length $buf;

	if ($ENV{REQUEST_METHOD} eq "HEAD") {
		printHeader("200 OK");
		print "Content-Length: 0\r\n",
		      "\r\n";

#	} elsif ($head[0] =~ /^Location/) {
#		print302("HTTP/1.0 302 Moved Temporarily");

	} elsif ($range) {
		my $all = length $buf;
		$buf = substr $buf, $range;
		$len = length $buf;
		printHeader("206 Partial content");
		print "Content-Range: bytes $range-" . ($all-1)  . "/$all\r\n",
		      "Content-Length: $len\r\n";

	} elsif ($gzip) {
		my $buf_gzip = Shingetsu::Extern::xgzip($buf);
		if (length($buf_gzip) < length($buf)) {
			$buf = $buf_gzip;
			$len = length $buf;
			printHeader("200 OK");
			print "Content-Encoding: gzip\r\n",
			      "Content-Length: $len\r\n";
		} else {
			printHeader("200 OK");
			print "Content-Length: $len\r\n";
		}

	} else {
		printHeader("200 OK");
		print "Content-Length: $len\r\n";
	}
	print @head;
	print "\r\n";

	if ($ENV{REQUEST_METHOD} ne "HEAD") {
		my $i;
		alarm $timeout;
		for ($i=0; $i<$len; $i+=$bufsize) {
			print substr $buf, $i, $bufsize;
			alarm $timeout;
		}
		alarm 0;
	}
}

#
# print file content
#
sub printFile {
	my($file) = @_;
	$file =~ s|^/+||;
	$file =~ /\.([^.]+)$/;
	my $suffix = $1;
	$suffix = "txt" unless (defined $suffix);
	my $type = $Shingetsu::Config::mimeType{$suffix};
	my $filedir = $Shingetsu::Config::filedir;
	$type = "text/plain" unless (defined $type);
	if ((-f "$filedir/$file")&&(open IN, "$filedir/$file")) {
		printHeader("200 OK");
		alarm $timeout;
		print "Content-Type: $type\r\n",
		      "Content-Length: ", (-s "$filedir/$file"), "\r\n",
		      "\r\n";
		my $buf;
		while (read IN, $buf, $bufsize) {
			print $buf;
			alarm $timeout;
		}
		alarm 0;
	} else {
		printHeader("404 Not Found");
		alarm $timeout;
		print "\r\nBad arguments or No data.\n",
		      "Try later.\n";
		alarm 0;
	}
}
